@@ -1,12 +1,10 @@
 from __future__ import annotations
 
 import os
-from typing import Dict, Any
+from typing import Dict, Any as Value
 
 from .exceptions import NoSuchEnvironmentVariableError
 
-Value = Any
-
 
 class Environment:
     def __init__(self, **env):
