@@ -1,12 +1,10 @@
 from __future__ import annotations
 
-from typing import Dict, List, Any, FrozenSet
+from typing import Dict, List, Any as Value
 
 from .environment import Environment
 from .exceptions import NoSuchVariableError, NoSuchEnvironmentVariableError
-from .util import Singleton, select, substitute, cachedmethod
-
-Value = Any
+from .util import Singleton, select, substitute, empty, cached_method
 
 
 class Node:
@@ -20,10 +18,7 @@ class Node:
         return self.parent.get_let(name)
 
     def evaluate(self, env: Environment) -> Value:
-        pass
-
-    def requires(self) -> FrozenSet[str]:
-        return frozenset()
+        return None
 
 
 class Empty(Node, metaclass=Singleton):
@@ -41,7 +36,7 @@ class Constant(Node):
         super().__init__(parent)
         self.value = value
 
-    def evaluate(self, env: Dict[str, Any]) -> Value:
+    def evaluate(self, env: Dict[str, Value]) -> Value:
         return self.value
 
 
@@ -56,15 +51,6 @@ class GetLet(Node):
             result = select(result, *self.path[1:])
         return result
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = {self.path[0]}
-        try:
-            result.update(self.get_let(self.path[0]).requires())
-        except NoSuchVariableError:
-            pass
-        return frozenset(result)
-
 
 class GetDyn(Node):
     def __init__(self, parent: Node, path: List[str]):
@@ -77,10 +63,6 @@ class GetDyn(Node):
             result = select(result, *self.path[1:])
         return result
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        return frozenset([f"${self.path[0]}"])
-
 
 class GetEnvVar(Node):
     def __init__(self, parent: Node, name: str):
@@ -90,9 +72,6 @@ class GetEnvVar(Node):
     def evaluate(self, env: Environment) -> Value:
         return env.get_var(self.name)
 
-    def requires(self) -> FrozenSet[str]:
-        return frozenset([self.name])
-
 
 class Let(Node):
     def __init__(self, parent: Node):
@@ -114,23 +93,6 @@ class Let(Node):
             return super().get_let(name)
         return self.let[name]
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set(self.body.requires())
-        for node in self.env.values():
-            result.update(node.requires())
-        for name in self.env.keys():
-            try:
-                result.remove(f"${name}")
-            except KeyError:
-                pass
-        for name in self.let.keys():
-            try:
-                result.remove(name)
-            except KeyError:
-                pass
-        return frozenset(result)
-
 
 class Else(Node):
     def __init__(self, parent: Node):
@@ -139,43 +101,17 @@ class Else(Node):
         self.otherwise: Node = Empty()
 
     def evaluate_body(self, env: Environment):
-        # noinspection PyBroadException
         try:
             return self.body.evaluate(env)
-        except (NoSuchEnvironmentVariableError, ArithmeticError):
+        except (NoSuchEnvironmentVariableError, ArithmeticError, ValueError, TypeError):
             return None
 
     def evaluate(self, env: Environment) -> Value:
         result = self.evaluate_body(env)
-        if self.empty(result):
+        if empty(result):
             result = self.otherwise.evaluate(env)
         return result
 
-    @staticmethod
-    def empty(result) -> bool:
-        if isinstance(result, (bool, list, dict, str)):
-            return not result
-        return result is None
-
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set(self.body.requires())
-
-        def is_env_var(name: str) -> bool:
-            return any(char.isupper() for char in name)
-
-        def is_env(name: str) -> bool:
-            return name[:1] == '$'
-
-        if not any(is_env(x) or is_env_var(x) for x in result):
-            body_result = self.body.evaluate(Environment())
-            if self.empty(body_result):
-                return self.otherwise.requires()
-            else:
-                return self.body.requires()
-        result.update(self.otherwise.requires())
-        return frozenset(result)
-
 
 class Object(Node):
     def __init__(self, parent: Node):
@@ -188,13 +124,6 @@ class Object(Node):
             result[key] = node.evaluate(env)
         return result
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set()
-        for node in self.children.values():
-            result.update(node.requires())
-        return frozenset(result)
-
 
 class Sequence(Node):
     def __init__(self, parent: Node):
@@ -207,13 +136,6 @@ class Sequence(Node):
             result.append(element.evaluate(env))
         return result
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set()
-        for node in self.elements:
-            result.update(node.requires())
-        return frozenset(result)
-
 
 class FunctionApplication(Node):
     def __init__(self, parent: Node, function):
@@ -228,13 +150,6 @@ class FunctionApplication(Node):
         result = self.function(*args)
         return result
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set()
-        for node in self.args:
-            result.update(node.requires())
-        return frozenset(result)
-
 
 class Application(Node):
     def __init__(self, parent: Node, template: str):
@@ -242,24 +157,13 @@ class Application(Node):
         self.template = template
         self.kwargs: Dict[str, Node] = {}
 
+    @cached_method
     def evaluate(self, env: Environment) -> Value:
         my_env: Dict[str, Value] = {}
         for key, node in self.kwargs.items():
             my_env[key] = node.evaluate(env)
         return self.get_let(self.template).evaluate(env.with_env(my_env))
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set(self.get_let(self.template).requires())
-        for node in self.kwargs.values():
-            result.update(node.requires())
-        for name in self.kwargs.keys():
-            try:
-                result.remove(f"${name}")
-            except KeyError:
-                pass
-        return frozenset(result)
-
 
 class Each(Node):
     def __init__(self, parent: Node, source: str, target: str):
@@ -293,15 +197,44 @@ class Each(Node):
                 results.append(result)
         return results
 
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set(self.body.requires())
-        result.add(self.source)
-        try:
-            result.remove(self.target)
-        except KeyError:
-            pass
-        return frozenset(result)
+
+class When(Node):
+    def __init__(self, parent: Node):
+        super().__init__(parent)
+        self.when: Node = Empty()
+        self.then: Node = Empty()
+        self.else_: Node = Empty()
+
+    def evaluate(self, env: Environment) -> Value:
+        if not empty(self.when.evaluate(env)):
+            return self.then.evaluate(env)
+        else:
+            return self.else_.evaluate(env)
+
+
+class All(Node):
+    def __init__(self, parent: Node):
+        super().__init__(parent)
+        self.nodes: List[Node] = []
+
+    def evaluate(self, env: Environment) -> Value:
+        for node in self.nodes:
+            if empty(node.evaluate(env)):
+                return False
+        return True
+
+
+class Any(Node):
+    def __init__(self, parent: Node):
+        super().__init__(parent)
+        self.nodes: List[Node] = []
+
+    def evaluate(self, env: Environment) -> Value:
+        for node in self.nodes:
+            result = node.evaluate(env)
+            if not empty(result):
+                return result
+        return False
 
 
 class Format(Node):
@@ -316,14 +249,3 @@ class Format(Node):
             return self.get_let(key).evaluate(env)
 
         return substitute(self.value, subst)
-
-    @cachedmethod
-    def requires(self) -> FrozenSet[str]:
-        result = set()
-
-        def record(key: str) -> str:
-            result.add(key)
-            return ""
-
-        substitute(self.value, record)
-        return frozenset(result)
