@@ -5,11 +5,13 @@ from datetime import date, datetime
 from inspect import getattr_static
 
 import yaml
+from dezimal import Dec
 
 from .exceptions import MalformedEachError, MalformedNameError, NoParseError, NoSuchFunctionError
+from .expressions import parse_expression
 from .functions import Functions
 from .nodes import *
-from .util import merge, Dec
+from .util import merge
 
 
 # noinspection PyMethodMayBeStatic
@@ -17,14 +19,13 @@ class Parser:
 
     def __init__(self):
         self.name_regex = "^[a-z]([_-]?[a-z0-9])*$"
-        self.env_var_regex = "^[0-9_]*[A-Z][_A-Z0-9]*$"
         self.path = []
 
     def check_name(self, name):
         if not re.match(self.name_regex, name):
             raise MalformedNameError(name=name, expected=self.name_regex)
 
-    def parse_node(self, obj: Any, parent: Node) -> Node:
+    def parse_node(self, obj: Value, parent: Node) -> Node:
         if isinstance(obj, list):
             return self.parse_sequence(obj, parent)
         if isinstance(obj, (type(None), bool, int, float, str, Dec, date, datetime)):
@@ -41,29 +42,34 @@ class Parser:
                 with open(include) as f:
                     docs.append(yaml.safe_load(f))
             obj = merge(*docs)
+        key_set = set(obj.keys())
+        if key_set == {'::all'}:
+            return self.parse_all(obj['::all'], parent)
+        if key_set == {'::any'}:
+            return self.parse_any(obj['::any'], parent)
+        if key_set == {'::when', '::then'} or key_set == {'::when', '::then', '::else'}:
+            return self.parse_conditional(obj, parent)
+        for keyword in ['::all', '::any', '::when', '::then']:
+            if keyword in key_set:
+                raise NoParseError()
         if '::else' in obj:
             return self.parse_else(obj, parent)
         if '::let' in obj:
             return self.parse_let(obj, parent)
         nodes = []
         remaining = {}
-        if '::ref' in obj:
-            ref = obj['::ref']
-            if isinstance(ref, str) and re.match(self.env_var_regex, ref):
-                nodes.append(self.parse_get_env_var(obj, parent))
-            elif isinstance(ref, str) and ref[:1] == '$' or isinstance(ref, list) and ref[0][:1] == '$':
-                nodes.append(self.parse_get_dyn(obj, parent))
-            else:
-                nodes.append(self.parse_get_let(obj, parent))
+        # TODO: decide - if ::when or ::then are here then NoParse - maybe?
         for key, value in obj.items():
-            if key[:8] == "::format":
+            if '::get' in obj:
+                nodes.append(parse_expression(obj['::get'], parent))
+            elif key[:8] == "::format":
                 nodes.append(self.parse_format(value, parent))
             elif key[:6] == "::call":
                 nodes.append(self.parse_application(key, value, parent))
             elif key[:6] == "::each":
                 nodes.append(self.parse_each(key, value, parent))
-            elif key[:2] == "::":
-                if key not in {"::ref", "::call", "::each", "::format"}:
+            elif len(key) > 2 and key[:2] == "::":
+                if key not in {"::get", "::call", "::each", "::format"}:
                     nodes.append(self.parse_function_application(key, value, parent))
             else:
                 remaining[key] = value
@@ -82,32 +88,11 @@ class Parser:
     def parse_constant(self, obj, parent: Node) -> Node:
         return Constant(parent, obj)
 
-    def parse_get_let(self, obj, parent: Node) -> Node:
-        path = obj['::ref']
-        if isinstance(path, str):
-            path = path.split(".")
-        if not isinstance(path, list):
-            raise NoParseError()
-        self.check_name(path[0])
-        return GetLet(parent, path)
-
-    def parse_get_dyn(self, obj, parent: Node) -> Node:
-        path = obj['::ref']
-        if isinstance(path, str):
-            path = path.split(".")
-        if not isinstance(path, list):
-            raise NoParseError()
-        path[0] = path[0][1:]  # remove leading dollar sign
-        self.check_name(path[0])
-        return GetDyn(parent, path)
-
-    def parse_get_env_var(self, obj, parent: Node) -> Node:
-        path = obj['::ref']
-        return GetEnvVar(parent, path)
-
-    def parse_let(self, obj: Any, parent: Node) -> Node:
+    def parse_let(self, obj: Value, parent: Node) -> Node:
         node = Let(parent)
         remainder = {}
+        if not isinstance(obj, dict):
+            raise NoParseError()
         for key, value in obj.items():
             if key == '::let':
                 for var_key, var_value in value.items():
@@ -123,7 +108,31 @@ class Parser:
         node.body = self.parse_node(remainder, node)
         return node
 
-    def parse_else(self, obj: Any, parent: Node) -> Node:
+    def parse_conditional(self, obj, parent: Node) -> Node:
+        node = When(parent)
+        node.when = self.parse_node(obj['::when'], node)
+        node.then = self.parse_node(obj['::then'], node)
+        if '::else' in obj:
+            node.else_ = self.parse_node(obj['::else'], parent)
+        return node
+
+    def parse_all(self, obj, parent: Node) -> Node:
+        if not isinstance(obj, list):
+            raise NoParseError()
+        node = All(parent)
+        for item in obj:
+            node.nodes.append(self.parse_node(item, node))
+        return node
+
+    def parse_any(self, obj, parent: Node) -> Node:
+        if not isinstance(obj, list):
+            raise NoParseError()
+        node = Any(parent)
+        for item in obj:
+            node.nodes.append(self.parse_node(item, node))
+        return node
+
+    def parse_else(self, obj: Value, parent: Node) -> Node:
         node = Else(parent)
         remainder = {}
         for key, value in obj.items():
